/*
The BYOND MUD Engine
Author: Dan (dan@dantom.com)
Version: 1.0 (08/01/99)

A library to quickly generate online inhabitable "rooms".  Consult mud.html
for detailed information.
*/

#include "room.dm"
#include "action.dm"
#include "object.dm"
#include "designer.dm"

//Make one global instance of /Mud called mud.
//The rest of the code assumes its existance.
var/Mud/mud = new()

world/New()
	mud.Read() //load the world

Mud/var
	savefile/File
	repop_period = 1000

	roomtypes[]
	actiontypes[]

	//all "population" types (including run-time additions)
	poptypes[0]

	//population types from the source code
	basic_poptypes[]

	dirty //flag indicating if changes to world need to be saved

Mud/proc
	//stuff you might want to override
	WelcomePlayer()
	WelcomeNewPlayer()

	//stuff you might need to call
	GotoStart(P)      //move player to the first room
	StorePlayer(P)    //write player savefile and destroy mob
	SavePlayer(P)     //write player savefile
	RemovePlayer(key) //destroy player savefile
	LocateRoom(roomid)

	//stuff you probably do not care about because it all works automatically
	Login()
	Logout()
	LoadPlayer(P)
	ExportPlayer(P,password,address)
	ImportPlayer(password,address)
	SavedPlayers()
	PlayerIsSaved(P)
	GetSaveFile()
	GetPlayerSaveFile(P,clearit)
	LocatePopType(poptypeid)
	NewRoomId()
	RepopTick()

Mud/New()
	//don't include action objects in the list of population types
	roomtypes = typesof(/room)
	actiontypes = typesof(/action)
	basic_poptypes = typesof(/obj,/mob)

Mud/RepopTick()
	var/room/a
	var/count = 1
	spawn for()
		count++
		for(a)
			if((count % a.repop_delay) == 0) a.Repop()
		sleep(repop_period)

Mud/GotoStart(mob/P)
	return P.Move(LocateRoom(1))

Mud/WelcomeNewPlayer()
	usr << "Welcome, [usr.name]."
Mud/WelcomePlayer()
	usr << "Welcome back, [usr.name]."

Mud/Login()
	if(!usr.loc)
		if(!LoadPlayer(usr))
			if(!usr) return //rejected or re-routed to another server
			WelcomeNewPlayer()
		else WelcomePlayer()

		//when running directly in the client, automatically give designer access
		if(!usr.IsDesigner())
			if(!usr.client.address || world.GetConfig("admin",usr.ckey))
				new/obj/designer_genes(usr)

		if(!usr.loc) GotoStart(usr)
	return 1

Mud/Logout()
	mud.StorePlayer(usr)

mob/Login()
	return mud.Login()

mob/Logout()
	if(key) //disconnect vs. moving from one mob to another
		mud.Logout()
	return ..()

Mud/GetSaveFile()
	if(!File) File = new("mud.sav")
	File.cd = "/"
	return File

Mud/GetPlayerSaveFile(mob/P,clearit)
	if(clearit) fdel("players/[P.ckey]")
	var/savefile/F = new("players/[P.ckey]")
	return F

Mud/StorePlayer(mob/P)
	SavePlayer(P)
	del P

world/Topic(topic,address)
	if(findtext(topic,"player.",1,8))
		return mud.ImportPlayer(copytext(topic,8),address)
	else ..()

Mud/ImportPlayer(password,address)
	//check list of registered imports for address and password
	var/action/import/O
	for(O)
		if((O.address == address || !O.address) && (O.password == password || !O.password))
			. = 1
			break;
	if(!.)
		world.log << "Refused import from '[address]' with password '[password]'."
		return

	var/savefile/F = new(world.Import())
	var/mob/P = new world.mob
	P.Read(F)
	P.key = F["key"] //make sure key gets restored
	StorePlayer(P)
	return 1

Mud/ExportPlayer(mob/P,password,address)
	var/savefile/F = new()

	P.Write(F)       //write player to temporary save file
	F["key"] = P.key //make sure key gets written

	//The server-to-server topic is in the form: player.<password>

	var/srv_address = copytext(address,1,findtext(address,"?"))
	srv_address = addtext(srv_address,"?player.",password)

	. = world.Export(srv_address,F)  //send the save file

	del(F)
	if(.)
		SavePlayer(P)
		F = GetPlayerSaveFile(P)
		F["export"] << address //mark this player as "exported"
		P << link(address)

client/New(topic)
	if(topic == "new_player") //player can use this to force export link to be ignored
		mud.RemovePlayer(key)
	return ..()

Mud/SavedPlayers()
	return flist("players/")

Mud/PlayerIsSaved(mob/P)
	return (P.ckey in SavedPlayers())

Mud/RemovePlayer(key)
	return fdel("players/[ckey(key)]")
	//TODO: backup the removed save file so it can be manually restored

Mud/SavePlayer(mob/P)
	var/savefile/F = GetPlayerSaveFile(P,1)
	P.Write(F)  //call mob.Write() directly to avoid saving of type info
	//del F

Mud/LoadPlayer(mob/P)
	if(!PlayerIsSaved(P)) return 0
	var/savefile/F = GetPlayerSaveFile(P)
	var/address = F["export"]
	if(address)
		P << "Forwarding you to a different locality...\n(If you <em>really</em> want to login here, click <a href=[world.url]?new_player>here</a>.)"
		P << link(address)
		del(P)
		return 0
	P.Read(F)
	//del F
	return 1

world/Del()
	if(mud.dirty) mud.Write()
	return ..()

Mud/Write(savefile/F)
	var/room/a

	if(!F)
		F = GetSaveFile()
		dirty = 0

	F.dir.Remove("rooms") //clear all saved rooms
	F.cd = "rooms"
	for(a)
		if(a.roomid > 0) F[num2text(a.roomid)] << a
	F.cd = ".."
	F["poptypes"] << mud.poptypes

Mud/Read(savefile/F)
	var/room/r
	var/action/a
	var/roomid
	var/P

	if(!F) F = GetSaveFile()
	F.cd = "rooms"
	for(roomid in F.dir)
		F[roomid] >> r
		r.roomid = text2num(roomid)
	F.cd = ".."

	//create code-defined rooms if they were not in the save file
	for(roomid in typesof(/room))
		r = new roomid()
		if(!r.roomid || LocateRoom(r.roomid) != r) del(r)

	F["poptypes"] >> mud.poptypes
	if(!mud.poptypes) mud.poptypes = list()
	while(mud.poptypes.Remove(null)) //in case some types were undefined

	for(P in mud.basic_poptypes)     //add any new basic types
		if(!locate(P) in mud.poptypes)
			mud.poptypes.Add(new P)

	for(r) r.PostRead()
	for(a) a.PostRead()

	dirty = 0
	RepopTick() //start up the repop ticker

Mud/NewRoomId()
	var/room/a
	var/roomid = 1
	start_over:
	for(a)
		if(a.roomid == roomid)
			roomid++
			goto start_over
	return roomid

Mud/LocateRoom(roomid)
	var/room/a
	if(isnum(roomid))
		for(a)
			if(a.roomid == roomid) return a
	else
		a = locate(roomid)
		if(a && !a.roomid)
			a.roomid = NewRoomId()
			a.PostRead() //room was implicitly created
		return a

Mud/LocatePopType(poptypeid)
	if(isnum(poptypeid)) return poptypes[poptypeid]
	return locate(poptypeid,poptypes)
