/************************************************
                ROOM DEFINITION

  Rooms are the containers in which the game
  takes place.  They are linked together by
  exits and populated with objects which
  periodically regenerate (repop) when destroyed.
*************************************************/

room
	parent_type = /area
	var
		roomid

		room
			north_exit   //These hard-coded exits could all be handled with exit
			south_exit   //actions, but since they are used so much, it's nice
			east_exit    //to have as little overhead as possible.
			west_exit

		population[0]
		repop_delay = 2 //repop every other cycle (default is 2 so some rooms can repop faster)

		tmp/post_read   //flag to prevent multiple calls to PostRead when implicitely creating soft-coded areas

	proc
		LinkToDir(O,Dir)
		GetLinkToDir(Dir)
		LinkNorth(O)
		LinkSouth(O)
		LinkEast(O)
		LinkWest(O)
		AddPop(O)
		DelPop(O)
		Repop()
		Unpop(O)
		Desc()
		PostRead() //called after mud is loaded to do any final processing that could not be done in Read()

room
	LinkToDir(O,Dir)
		switch(Dir)
			if(NORTH) LinkNorth(O)
			if(SOUTH) LinkSouth(O)
			if(EAST) LinkEast(O)
			if(WEST) LinkWest(O)
			else
				world.log << "Unexpected linkage direction in [type]/LinkToDir([Dir],[O])"
	GetLinkToDir(Dir)
		switch(Dir)
			if(NORTH) return north_exit
			if(SOUTH) return south_exit
			if(EAST) return east_exit
			if(WEST) return west_exit

	LinkNorth(O)
		if(O) verbs.Add(/room/proc/north)
		else verbs.Remove(/room/proc/north)
		north_exit = O
	LinkSouth(O)
		if(O) verbs.Add(/room/proc/south)
		else verbs.Remove(/room/proc/south)
		south_exit = O
	LinkEast(O)
		if(O) verbs.Add(/room/proc/east)
		else verbs.Remove(/room/proc/east)
		east_exit = O
	LinkWest(O)
		if(O) verbs.Add(/room/proc/west)
		else verbs.Remove(/room/proc/west)
		west_exit = O
	DelPop(atom/movable/O)
		population.Remove(O)
		O.poploc = null
	AddPop(O)
		population.Add(O)
	VarIsEditable(V)
		switch(V)
			if("population","gender","visibility","roomid")
				return 0
		return ..()
	GetDesignerText()
		return "<a href=?\ref[src]DesignerInfo>[name || type]</a> ([roomid])"
	DesignerInfo()
		if(roomid < 0) usr << "roomid: [roomid] (hard-coded)"
		else usr << "roomid: [roomid]"
		..()

		var/count
		var/atom/movable/O
		for(O in population)
			if(count++)
				usr << ", \..."
			else
				usr << "population: \..."
			usr << "[O.GetDesignerText()]\..."
		if(count) usr << ""

		count = 0
		var/action/A
		for(A in src)
			if(count++)
				usr << ", \..."
			else
				usr << "actions: \..."
			usr << "[A.GetDesignerText()]\..."
		if(count) usr << ""


	Configure(property)
		switch(property)
			if("north_exit","south_exit","east_exit","west_exit")
				var/room/new_exit
				var/room/old_exit = vars[property]
				var/old_exit_id
				if(old_exit) old_exit_id = old_exit.roomid
				new_exit = input("Enter a room id (0 for none) or walk to the desired room and enter -1.",property,old_exit_id) as num|null
				if(isnull(new_exit)) return //canceled
				if(new_exit == -1) new_exit = usr.loc
				else if(new_exit == 0) new_exit = null
				else
					new_exit = mud.LocateRoom(new_exit)
					if(!new_exit)
						usr << "No such room."
						return

				var/new_dir
				switch(property)
					if("north_exit") new_dir = NORTH
					if("south_exit") new_dir = SOUTH
					if("east_exit") new_dir = EAST
					if("west_exit") new_dir = WEST

				LinkToDir(new_exit,new_dir)

				new_dir = turn(new_dir,180)
				if(new_exit)
					if(alert("Make this a bidirectional link between [name] and [new_exit.name]?","Bidirectional","Yes","No") == "Yes")
						new_exit.LinkToDir(src,new_dir)
				else if(old_exit && old_exit.GetLinkToDir(new_dir) == src)
					if(alert("Remove the bidirection link from [old_exit.name] to [name]?","Remove bidirection exit","Yes","No") == "Yes")
						old_exit.LinkToDir(null,new_dir)
				return 1
			else return ..()
	proc/LinkExitName(exit_name,room/dest)
		switch(exit_name)
			if("north_exit") LinkNorth(dest)
			if("south_exit") LinkSouth(dest)
			if("east_exit") LinkEast(dest)
			if("west_exit") LinkWest(dest)

	Repop()
		var
			i; P; N
		for(i = population.len,i>0,i--)
			P = population[i]
			if(P:poptype) continue //already populated (and still kicking)
			N = P:Clone(src)
			N:Repop()
			population[i] = N //replace poptype with new instance (when instance dies it will change it back)
	Unpop(O)
		//swap pop type back into population list
		var/i = population.Find(O)
		population[i] = O:poptype
	Desc()
		usr << name
		if(desc) usr << desc
		usr << ""
		if(north_exit) usr << "north [north_exit]"
		if(south_exit) usr << "south [south_exit]"
		if(east_exit) usr << "east [east_exit]"
		if(west_exit) usr << "west [west_exit]"
		var/action/exit/e
		for(e in src)
			if(e.blind) continue
			usr << "[e.name] [e.dest]"

		usr << ""
		var/O
		for(O in src)
			if(!O:visibility) continue
			usr << O

	Stat()
		stat(name)
		stat(desc)
		stat(null) //blank line
		if(north_exit) stat("north",north_exit)
		if(south_exit) stat("south",south_exit)
		if(east_exit) stat("east",east_exit)
		if(west_exit) stat("west",west_exit)

		var/action/exit/e
		for(e in src)
			if(e.blind || e.hidden) continue
			stat(e.name,e.dest)
		var/action/a
		for(a in src)
			e = a
			if(istype(e) && !e.blind) continue
			if(a.hidden) continue
			stat(a)

		var/O
		stat(null) //blank line
		for(O in src)
			if(!O:visibility) continue
			stat(O)

	Write(savefile/F)
		F["name"] << name
		F["desc"] << desc
		if(north_exit) F["north"] << north_exit:roomid
		if(south_exit) F["south"] << south_exit:roomid
		if(east_exit) F["east"] << east_exit:roomid
		if(west_exit) F["west"] << west_exit:roomid

		//write the action list
		var/action/A
		var/actions[0]
		for(A in src) actions.Add(A)
		if(actions.len) F["contents"] << actions

		//write the population list (as population ids)
		if(population.len)
			var/popids[0]
			var/P
			for(P in population)
				if(P:poptype) P = P:poptype
				popids.Add(mud.poptypes.Find(P))
			F["population"] << popids
		if(repop_delay != initial(repop_delay)) F["repop_delay"] << repop_delay


	Read(savefile/F)
		F["name"] >> name
		F["desc"] >> desc
		F["north"] >> north_exit
		F["south"] >> south_exit
		F["east"] >> east_exit
		F["west"] >> west_exit
		F["contents"] >> contents
		F["population"] >> population
		F["repop_delay"] >> repop_delay
		if(!repop_delay) repop_delay = 2

	PostRead()
		if(post_read) return
		post_read = 1

		if(north_exit) LinkNorth(mud.LocateRoom(north_exit))
		if(south_exit) LinkSouth(mud.LocateRoom(south_exit))
		if(east_exit) LinkEast(mud.LocateRoom(east_exit))
		if(west_exit) LinkWest(mud.LocateRoom(west_exit))

		var/i
		//convert population ids to population types
		if(!population) population = list()
		else
			for(i=population.len,i,i--)
				var/pop = mud.LocatePopType(population[i])
				if(!pop) world.log << "warning: unable to load pop type [population[i]]"
				population[i] = pop
			while(population.Remove(null)) {} //remove any failed cases

//verbs to be added dynamically to rooms (defined as procs so they don't show up by default)
room/proc
	DoAction(name)
		var/action/a
		for(a in src)
			if(a.name == name) a.Action()
	north()
		if(usr.Move(north_exit))
			src << "[usr] leaves to the north."
			oview(usr) << "[usr] arrives from the south."
		else DoAction("north")
	south()
		if(usr.Move(south_exit))
			src << "[usr] leaves to the south."
			oview(usr) << "[usr] arrives from the north."
		else DoAction("south")
	east()
		if(usr.Move(east_exit))
			src << "[usr] leaves to the east."
			oview(usr) << "[usr] arrives from the west."
		else DoAction("east")
	west()
		if(usr.Move(west_exit))
			src << "[usr] leaves to the west."
			oview(usr) << "[usr] arrives from the east."
		else DoAction("west")

//translate direction keys to exit procs
client
	North()
		if(isarea(usr.loc)) usr.loc:north()
		else ..()
	South()
		if(isarea(usr.loc)) usr.loc:south()
		else ..()
	East()
		if(isarea(usr.loc)) usr.loc:east()
		else ..()
	West()
		if(isarea(usr.loc)) usr.loc:west()
		else ..()

client/Click(O)
	if(isarea(O))
		var/room/r = usr.loc
		if(r.north_exit == O) return r.north()
		if(r.south_exit == O) return r.south()
		if(r.east_exit == O) return r.east()
		if(r.west_exit == O) return r.west()

		var/action/exit/e
		for(e in r)
			if(e.dest == O) return e.Action()
	else ..()

